#!/usr/bin/env python
# -*- coding: utf8 -*-
'''
Show a ports dependence tree.

Three styles:
0 - text only
1 - single line pseudographics (default)
2 - double line pseudographics
A node with descendents and back reference has extra symbol indicating this
'''

from subprocess import Popen, PIPE
from optparse import OptionParser
import networkx as nx
import fnmatch
import os
import cPickle as pickle

# one of the next two lines is needed for graphical output
#import pylab as P
#import matplotlib.pyplot as P

#DEPENDS_PREFIXES = ('FETCH', 'EXTRACT', 'PATCH', 'BUILD', 'LIB', 'RUN') 

class Net():
    'Show a network as a tree with back references using pseudo-graphics'
    none = (u']', u'⊢', u'⊫', u'⊢', u'⊫')
    #parent = (u'-', u'─', u'═', u'─', u'═')
    child = (u'+', u'┌', u'╔', u'┌', u'╔')
    both = (u'+', u'┬', u'╦', u'', u'')
    ref = (u' -> ', u' ─> ', u' ═> ', u' → ', u' ⇒ ')
    now_then = (u'+', u'├', u'╠', u'├', u'╠')
    then = (u'|', u'│', u'║', u'│', u'║')
    now = (u'`', u'└', u'╚', u'└', u'╚')
    never = (u' ', u' ', u' ', u' ', u' ')

    def __init__(self, net=nx.DiGraph()):
        self.net = net
        self.line = 0
        self.style = 1
        self.number_all = False
        self.shown = {}
        self.ignore = len("/usr/ports/")

    def name(self, index):
        'Return names on a node determined by index'
        return str(index[self.ignore:])
##        try:
##            return "%s: %s" % (str(index), self.net[index])
##        except StandardError:
##            return str(index)

    def deeper1(self, parent_is_last, has_child):
        'Prepare addition to prefix for deeper tree traversing'
        if parent_is_last is None and has_child:
            return self.child[self.style] # '┌'
        if parent_is_last is None and not has_child:
            return self.none[self.style] # '⊢'
        if parent_is_last and has_child:
            return self.now[self.style]+self.both[self.style] # u'└┬'
        if parent_is_last and not has_child:
            return self.now[self.style] # u'└'
        if not parent_is_last and has_child:
            return self.now_then[self.style]+self.both[self.style] # u'├┬'
        if not parent_is_last and not has_child:
            return self.now_then[self.style] # u'├'

    def deeper2(self, parent_is_last):
        'Prepare addition to prefix for deeper tree traversing'
        if parent_is_last is None:
            return u'' # u''
        if parent_is_last:
            return self.never[self.style] # u' '
        if not parent_is_last:
            return self.then[self.style] # u'│'

    def show_descendants(self, index=0, prefix=u"", last=None):
        'Print a directed net- or tree-like object as a pseudo-graphical tree.'
        children = self.net.neighbors(index)
        if index in list(self.shown.keys()):
            result = u''
            if self.number_all:
                result = u'%d' % self.line
                self.line += 1
            result += u'\t%s%s%s%s%d' % \
                (prefix, self.deeper1(last, children), self.name(index),
                 self.ref[self.style], self.shown[index]) #' → '
            return result
        result = u'%d\t%s%s%s' % \
            (self.line, prefix, self.deeper1(last, children), self.name(index))
        self.shown[index] = self.line
        self.line += 1
        new_prefix = prefix + self.deeper2(last)
        if children:
            for child in children[:-1]:
                result = u'\n'.join([result,
                    self.show_descendants(child, new_prefix, last=False)])
            result = u'\n'.join([result,
                self.show_descendants(children[-1], new_prefix, last=True)])
        return result

    def show(self, index=None, style=None, number_all=None):
        'Show whole graph'
        if style is not None:
            self.style = style
        if number_all is not None:
            self.number_all = number_all
        if index is not None:
            return self.show_descendants(index)
        else:
            return u'\n'.join([self.show_descendants(ind)
                                for ind in self.net.nodes() 
                                    if ind not in list(self.shown.keys())])

def parse_args():
    'Set a parser up and parse the args.'
    parser = OptionParser()
    parser.add_option("-f", "--fetch-depends", dest="fetch", 
                    action="store_true", help="Use fetch dependences for tree")
    parser.add_option("-e", "--extract-depends", dest="extract",
                   action="store_true", help="Use extract dependences for tree")
    parser.add_option("-p", "--patch-depends", dest="patch",
                    action="store_true", help="Use patch dependences for tree")
    parser.add_option("-b", "--build-depends", dest="build",
                    action="store_true", help="Use build dependences for tree")
    parser.add_option("-l", "--lib-depends", dest="lib",
                    action="store_true", help="Use lib dependences for tree")
    parser.add_option("-r", "--run-depends", dest="run",
                    action="store_true", help="Use run dependences for tree")
    parser.add_option("-n", "--number-all", dest="number_all",
                    action="store_true",
                    help="Number all lines, even with repeated port names")
    parser.add_option("-s", "--style", default=1, metavar="NUM",
                      dest="style", help="set style to NUM (0...4)")
    parser.add_option("-A", "--all-ports", dest="all_ports",
                    action="store_true", help="Use all ports (this is slow!)")
    parser.add_option("-R", "--reverse", dest="reverse",
                    action="store_true", help="Also show reversed tree")
    parser.add_option("-O", "--others", dest="others",
                    action="store_true", help="Show all ports")
    parser.add_option("-S", "--save", dest="save",
                    action="store_true", help="Save the traversed network")
    parser.add_option("-U", "--unsave", dest="unsave",
                    action="store_true", help="Use the saved traversed network")

    options, args = parser.parse_args()
    options.all = True not in (options.fetch, options.extract, options.patch, \
                options.build, options.lib, options.run)
    options.style=eval('int(%s)' % str(options.style))

    if not args or args == []:
        args = ['.']

    return options, args

def get_depends(var, options):
    "Get all depends needed according to options"
    fetch_depend, extract_depend, patch_depend, build_depend, \
        lib_depend, run_depend, dummy = var.split('\n')
    all_depend = []
    if options.all or options.fetch:
        all_depend.append(fetch_depend)
    if options.all or options.extract:
        all_depend.append(extract_depend)
    if options.all or options.patch:
        all_depend.append(patch_depend)
    if options.all or options.build:
        all_depend.append(build_depend)
    if options.all or options.lib:
        all_depend.append(lib_depend)
    if options.all or options.run:
        all_depend.append(run_depend)
    return all_depend

def traverse(options, port, port_tree):
    "Extend port_tree by port and its descendents"
    port_tree.add_node(port, visited=False)
    stop = False
    while not stop:
        stop = True
        for port in port_tree.nodes(data=False)[:]:
            if port_tree.node[port]['visited']:
                continue
            port_tree.node[port]['visited'] = True
            var, err = Popen(('make', '-C', port,
                            '-V', 'FETCH_DEPENDS', 
                            '-V', 'EXTRACT_DEPENDS', 
                            '-V', 'PATCH_DEPENDS', 
                            '-V', 'BUILD_DEPENDS', 
                            '-V', 'LIB_DEPENDS', 
                            '-V', 'RUN_DEPENDS'), 
                        stdout=PIPE, stderr=PIPE).communicate()
            if err:
                print err
                continue
            for depend in get_depends(var, options):
                if depend in (None, ''):
                    continue
                for dep in depend.split():
                    dep = dep.split(':')[1]
                    if dep not in port_tree.nodes():
                        port_tree.add_node(dep, visited=False)
                        stop = False
                    port_tree.add_edge(port, dep) # repeating an edge is OK

def main():
    "Do the job"
    options, args = parse_args()

    if options.unsave:
        savefile = open("pickle", "rb")
        port_tree = pickle.load(savefile)
        savefile.close()
    else:
        port_tree = nx.DiGraph()
        if options.all_ports:
            for category in fnmatch.filter(os.listdir("/usr/ports"), "[a-z]*"):
                cat = os.path.join("/usr/ports", category)
                if not os.path.isdir(cat):
                    continue
                print cat
                for port in os.listdir(cat):
                    port1 = os.path.join(cat, port)
                    if not os.path.isdir(port1):
                        continue
                    traverse(options, port1, port_tree)
        else:
            for port in args:
                traverse(options, port, port_tree)
    net = Net(port_tree)
    print "Depending on"
    for port in args:
        print net.show(port, options.style, number_all=options.number_all)
    if options.others:
        print net.show(style=options.style, number_all=options.number_all)
    if options.reverse:
        print "\nRequired by:"
        net = Net(port_tree.reverse())
        for port in args:
            print net.show(port, options.style, number_all=options.number_all)
        if options.others:
            print net.show(style=options.style, number_all=options.number_all)

    if options.save:
        savefile = open("pickle", "wb")
        pickle.dump(port_tree, savefile, -1) # save DiDraph, highest protocol
        savefile.close()
    # get graphical output with the following lines (plus an import above)
    #nx.draw(port_tree)
    #P.draw()
    #P.show()

if __name__ == "__main__":
    main()
